"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const vscode = require("vscode");
class BufferBinding {
    constructor({ buffer, isHost, didDispose }) {
        this.buffer = buffer;
        this.isHost = isHost;
        this.didDispose = didDispose;
    }
    dispose() {
        this.disposed = true;
    }
    isDisposed() {
        return this.disposed;
    }
    getText() {
        if (typeof this.onGetText === "function") {
            return this.onGetText();
        }
        return null;
    }
    setBufferProxy(bufferProxy) {
        this.bufferProxy = bufferProxy;
    }
    setText(text) {
        fs.writeFileSync(this.buffer.uri.fsPath, text);
    }
    setEditor(editor) {
        this.editor = editor;
    }
    updateText(textUpdates) {
        return this.editor.edit(builder => {
            for (let i = textUpdates.length - 1; i >= 0; i--) {
                const textUpdate = textUpdates[i];
                builder.replace(this.createRange(textUpdate.oldStart, textUpdate.oldEnd), textUpdate.newText);
            }
        }, { undoStopBefore: false, undoStopAfter: true });
    }
    traverse(start, distance) {
        if (distance.row === 0) {
            return { row: start.row, column: start.column + distance.column };
        }
        else {
            return { row: start.row + distance.row, column: distance.column };
        }
    }
    insert(position, text) {
        console.log("buffer insert pos:" + position + " text: " + text);
        if (typeof this.onInsert === "function") {
            this.onInsert(position, text);
        }
        return [position, position, text];
    }
    delete(startPosition, extent) {
        console.log("buffer delete start pos:" + startPosition + " extent: " + extent);
        if (typeof this.onDelete === "function") {
            this.onDelete(startPosition, extent);
        }
        const endPosition = this.traverse(startPosition, extent);
        return [startPosition, endPosition, ''];
    }
    createRange(start, end) {
        return new vscode.Range(new vscode.Position(start.row, start.column), new vscode.Position(end.row, end.column));
    }
    onDidChangeBuffer(changes) {
        this.bufferProxy.onDidChangeBuffer(changes.map(change => {
            const { start, end } = change.range;
            return {
                oldStart: { row: start.line, column: start.character },
                oldEnd: { row: end.line, column: end.character },
                newText: change.text
            };
        }));
    }
    requestSavePromise() {
        return new Promise(() => {
            this.bufferProxy.requestSave();
        });
    }
    save() {
        this.buffer.save();
    }
}
exports.default = BufferBinding;
//# sourceMappingURL=BufferBinding.js.map