"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
class EditorBinding {
    constructor({ editor, portal, isHost }) {
        this.editor = editor;
        this.portal = portal;
        this.isHost = isHost;
        this.localSelectionMap = {};
        this.selectionsBySiteId = new Map();
        this.decorationBySiteId = new Map();
        this.localMarkerSelectionMap = new Map();
    }
    dispose() {
        this.disposed = true;
    }
    isDisposed() {
        return this.disposed;
    }
    onDidDispose(onDidDipose) {
        this.onDidDispose = onDidDipose;
    }
    setEditorProxy(editorProxy) {
        this.editorProxy = editorProxy;
    }
    updateSelectionsForSiteId(siteId, selectionUpdates) {
        console.log("updateSelectionsForSiteID: " + siteId);
        let selectionsForSite = this.localMarkerSelectionMap.get(siteId);
        const selectionMap = Object.assign(Object.assign({}, selectionsForSite), selectionUpdates);
        this.localMarkerSelectionMap.set(siteId, selectionMap);
        let selectionRanges = [];
        let cursorRanges = [];
        if (!selectionsForSite) {
            selectionsForSite = {};
            this.selectionsBySiteId[siteId] = selectionsForSite;
        }
        for (const selectionId in selectionUpdates) {
            const selectionUpdate = selectionUpdates[selectionId];
            if (selectionUpdate) {
                selectionsForSite[selectionId] = selectionUpdate;
                if (this.isCursor(selectionUpdate)) {
                    cursorRanges = cursorRanges.concat(this.convertTeletypeRange(selectionUpdate.range));
                }
                else {
                    if (selectionUpdate.tailed) {
                        const cursorRange = this.getCursorRangeFromSelection(selectionUpdate);
                        cursorRanges = cursorRanges.concat(this.convertTeletypeRange(cursorRange));
                    }
                    selectionRanges = selectionRanges.concat(this.convertTeletypeRange(selectionUpdate.range));
                }
            }
            else {
                delete selectionsForSite[selectionId];
            }
        }
        let siteDecoration = this.findSiteDecoration(siteId);
        this.updateDecorations(siteDecoration, cursorRanges, selectionRanges);
    }
    updateDecorations(siteDecoration, cursorRanges, selectionRanges) {
        const { cursorDecoration, selectionDecoration } = siteDecoration;
        this.editor.setDecorations(cursorDecoration, cursorRanges);
        this.editor.setDecorations(selectionDecoration, selectionRanges);
    }
    findSiteDecoration(siteId) {
        let siteDecoration = this.decorationBySiteId.get(siteId);
        if (!siteDecoration) {
            siteDecoration = this.createDecorationFromSiteId(siteId);
            this.decorationBySiteId.set(siteId, siteDecoration);
        }
        return siteDecoration;
    }
    isScrollNeededToViewPosition(position) {
    }
    updateTether(state, position) {
    }
    clearSelectionsForSiteId(siteId) {
        const siteDecoration = this.findSiteDecoration(siteId);
        this.updateDecorations(siteDecoration, [], []);
    }
    updateSelections(selections) {
        this.processSelections(selections);
        this.editorProxy.updateSelections(this.localSelectionMap);
    }
    processSelections(selections) {
        const currentSelectionKeys = Object.keys(this.localSelectionMap);
        const newSelectionsLength = selections.length;
        selections.forEach((selection, index) => {
            this.localSelectionMap[index] = {
                range: {
                    start: this.convertVSCodePosition(selection.start),
                    end: this.convertVSCodePosition(selection.end)
                },
                reversed: selection.isReversed,
            };
        });
        selections.forEach((selection, index) => {
            if (currentSelectionKeys.length > newSelectionsLength) {
                for (let index = newSelectionsLength; index < currentSelectionKeys.length; index += 1) {
                    this.localSelectionMap[index] = {
                        range: {
                            start: this.convertVSCodePosition(selection.start),
                            end: this.convertVSCodePosition(selection.end)
                        },
                        reversed: false,
                    };
                }
            }
        });
    }
    convertVSCodePosition(position) {
        return {
            column: position.character,
            row: position.line
        };
    }
    convertTeletypePosition(position) {
        return new vscode.Position(position.row, position.column);
    }
    convertTeletypeRange(range) {
        return new vscode.Range(this.convertTeletypePosition(range.start), this.convertTeletypePosition(range.end));
    }
    createDecorationFromSiteId(siteId) {
        const selectionDecorationRenderOption = {
            backgroundColor: `rgba(0,0,255,0.6)`
        };
        const { login: siteLogin } = this.portal.getSiteIdentity(siteId);
        const nameTagStyleRules = {
            position: 'absolute',
            top: '12px',
            padding: '0px 5px 0px 0px',
            display: 'inline-block',
            'z-index': 1,
            'border-radius': '20px',
            'font-size': '15px',
            'font-weight': 'bold'
        };
        const curosrDecorationRenderOption = {
            border: 'solid rgba(0,0,255,0.6)',
            borderWidth: '5px 5px 5px 5px',
            after: {
                contentText: siteLogin,
                backgroundColor: 'rgba(0,0,255,0.6)',
                color: 'rgba(192,192,192,30)',
                textDecoration: `none; ${this.stringifyCssProperties(nameTagStyleRules)}`
            }
        };
        const create = vscode.window.createTextEditorDecorationType;
        return {
            selectionDecoration: create(selectionDecorationRenderOption),
            cursorDecoration: create(curosrDecorationRenderOption)
        };
    }
    stringifyCssProperties(rules) {
        return Object.keys(rules)
            .map((rule) => {
            return `${rule}: ${rules[rule]};`;
        }).join(' ');
    }
    getCursorRangeFromSelection(selection) {
        const { range: { end, start } } = selection;
        if (selection.reversed) {
            return {
                start: start,
                end: start
            };
        }
        else {
            return {
                start: end,
                end: end
            };
        }
    }
    isCursor(selection) {
        const { start, end } = selection.range;
        return (start.column === end.column &&
            start.row === end.row);
    }
}
exports.default = EditorBinding;
//# sourceMappingURL=EditorBinding.js.map