"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require('assert');
const vscode = require("vscode");
const teletype_client_1 = require("@atom/teletype-client");
const BufferBinding_1 = require("./BufferBinding");
const EditorBinding_1 = require("./EditorBinding");
const os = require("os");
const path = require("path");
const fs = require("fs");
class PortalBinding {
    constructor({ client, portalId, editor }) {
        this.hostClosedPortal = false;
        this.hostLostConnection = false;
        this.leaveEvents = [];
        this.editorProxies = new Set();
        this.tetherPosition = null;
        this.activePositionsBySiteId = {};
        this.client = client;
        this.portalId = portalId;
        this.editor = editor;
        this.lastEditorProxyChangePromise = Promise.resolve();
        this.editorBindingsByEditorProxy = new Map();
        this.bufferBindingsByBufferProxy = new Map();
        this.bufferBindingsByBuffer = new Map();
        this.editorBindingsByEditor = new Map();
        this.editorProxiesByEditor = new WeakMap();
        this.tetherEditorProxyChangeCounter = 0;
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.portal = yield this.client.joinPortal(this.portalId);
                this.portal.setDelegate(this);
                vscode.window.showInformationMessage('Joined Portal with ID' + ' ' + this.portalId + ' ');
                this.registerWorkspaceEvents();
            }
            catch (error) {
                vscode.window.showErrorMessage('Unable to join the Portal with ID' + ' ' + this.portalId + ' ');
            }
        });
    }
    registerWorkspaceEvents() {
        vscode.workspace.onDidChangeTextDocument(this.onDidChangeTextDocument.bind(this));
        vscode.workspace.onWillSaveTextDocument(this.saveDocument.bind(this));
        vscode.window.onDidChangeTextEditorSelection(this.triggerSelectionChanges.bind(this));
    }
    onDidChangeTextDocument(event) {
        if (this.bufferBindingsByBuffer) {
            const bufferBinding = this.bufferBindingsByBuffer.get(event.document);
            if (bufferBinding) {
                bufferBinding.onDidChangeBuffer(event.contentChanges);
            }
        }
    }
    saveDocument(event) {
        if (this.bufferBindingsByBuffer) {
            const bufferBinding = this.bufferBindingsByBuffer.get(event.document);
            if (bufferBinding) {
                event.waitUntil(bufferBinding.requestSavePromise());
            }
        }
    }
    triggerSelectionChanges(event) {
        const editorBinding = this.editorBindingsByEditor.get(event.textEditor);
        if (editorBinding) {
            editorBinding.updateSelections(event.selections);
        }
    }
    dispose() {
        this.disposed = true;
    }
    isDisposed() {
        return this.disposed;
    }
    hostDidClosePortal() {
        this.hostClosedPortal = true;
    }
    hasHostClosedPortal() {
        return this.hostClosedPortal;
    }
    hostDidLoseConnection() {
        this.hostLostConnection = true;
    }
    hasHostLostConnection() {
        return this.hostLostConnection;
    }
    addEditorProxy(editorProxy) {
        if (editorProxy && this.editorProxies) {
            if (typeof this.onAddEditorProxy === "function") {
                this.onAddEditorProxy(editorProxy);
            }
            console.log("addEditorProxy: " + editorProxy.bufferProxy.uri);
            if (!this.editorProxies.has(editorProxy)) {
                console.log('Cannot add the same editor proxy multiple times remove/add again');
                this.editorProxies.delete(editorProxy);
            }
            this.editorProxies.add(editorProxy);
        }
    }
    removeEditorProxy(editorProxy) {
        if (typeof this.onRemoveEditorProxy === "function") {
            this.onRemoveEditorProxy(editorProxy);
        }
        assert(this.editorProxies.has(editorProxy), 'Can only remove editor proxies that had previously been added');
        this.editorProxies.delete(editorProxy);
        if (this.tetherEditorProxy === editorProxy) {
            this.tetherEditorProxy = null;
            this.tetherEditorProxyChangeCounter++;
        }
    }
    editorProxyForURI(uri) {
        return Array.from(this.editorProxies).find((e) => e.bufferProxy.uri === uri);
    }
    getTetherEditorProxy() {
        return this.tetherEditorProxy;
    }
    getTetherBufferProxyURI() {
        return (this.tetherEditorProxy) ? this.tetherEditorProxy.bufferProxy.uri : null;
    }
    getEditorProxies() {
        return Array.from(this.editorProxies);
    }
    updateTether(state, editorProxy, position) {
        return __awaiter(this, void 0, void 0, function* () {
            if (editorProxy) {
                this.lastEditorProxyChangePromise = this.lastEditorProxyChangePromise.then(() => this.onUpdateTether(state, editorProxy, position));
            }
            console.log("updateTether: " + editorProxy.bufferProxy.uri);
            this.addEditorProxy(editorProxy);
            this.tetherState = state;
            if (editorProxy !== this.tetherEditorProxy) {
                this.tetherEditorProxy = editorProxy;
                this.tetherEditorProxyChangeCounter++;
            }
            this.tetherPosition = position;
        });
    }
    onUpdateTether(state, editorProxy, position) {
        return __awaiter(this, void 0, void 0, function* () {
            if (state === teletype_client_1.FollowState.RETRACTED) {
                const editor = yield this.findOrCreateEditorByEditorProxy(editorProxy);
            }
            else {
                this.editorBindingsByEditorProxy.forEach((editor_binding) => editor_binding.updateTether(state, undefined));
            }
            const editorBinding = this.editorBindingsByEditorProxy.get(editorProxy);
            if (editorBinding && position) {
                editorBinding.updateTether(state, position);
            }
        });
    }
    findOrCreateEditorByEditorProxy(editorProxy) {
        return __awaiter(this, void 0, void 0, function* () {
            let editor;
            let editorBinding = this.editorBindingsByEditorProxy.get(editorProxy);
            if (editorBinding) {
                editor = editorBinding.editor;
            }
            else {
                const { bufferProxy } = editorProxy;
                const buffer = yield this.findOrCreateBufferForBufferProxy(bufferProxy);
                const bufferBinding = this.bufferBindingsByBufferProxy.get(bufferProxy);
                yield vscode.workspace.openTextDocument(buffer.uri);
                console.log('find buffer, now show it');
                editor = yield vscode.window.showTextDocument(buffer);
                editorBinding = new EditorBinding_1.default({
                    editor,
                    portal: this.portal,
                    isHost: false
                });
                yield vscode.commands.executeCommand('workbench.action.keepEditor');
                if (bufferBinding) {
                    bufferBinding.setEditor(editor);
                }
                editorBinding.setEditorProxy(editorProxy);
                editorProxy.setDelegate(editorBinding);
                this.editorBindingsByEditorProxy.set(editorProxy, editorBinding);
                this.editorProxiesByEditor.set(editor, editorProxy);
                this.editorBindingsByEditor.set(editor, editorBinding);
                editorBinding.onDidDispose(() => {
                    this.editorProxiesByEditor.delete(editor);
                    this.editorBindingsByEditorProxy.delete(editorProxy);
                });
            }
            return editor;
        });
    }
    findOrCreateBufferForBufferProxy(bufferProxy) {
        return __awaiter(this, void 0, void 0, function* () {
            let buffer;
            let bufferBinding = this.bufferBindingsByBufferProxy.get(bufferProxy);
            if (bufferBinding) {
                buffer = bufferBinding.buffer;
            }
            else {
                const bufferURI = vscode.Uri.parse(`file://${path.join(os.tmpdir(), `/${this.portalId}/`, bufferProxy.uri)}`);
                yield require('mkdirp-promise')(path.dirname(bufferURI.fsPath));
                fs.writeFileSync(bufferURI.fsPath, '');
                buffer = yield vscode.workspace.openTextDocument(bufferURI);
                bufferBinding = new BufferBinding_1.default({
                    buffer,
                    isHost: false,
                    didDispose: () => this.bufferBindingsByBufferProxy.delete(bufferProxy)
                });
                bufferBinding.setBufferProxy(bufferProxy);
                bufferProxy.setDelegate(bufferBinding);
                this.bufferBindingsByBuffer.set(buffer, bufferBinding);
                this.bufferBindingsByBufferProxy.set(bufferProxy, bufferBinding);
            }
            return buffer;
        });
    }
    getTetherState() {
        return this.tetherState;
    }
    getTetherPosition() {
        return this.tetherPosition;
    }
    updateActivePositions(positionsBySiteId) {
        // this.sitePositionsComponent.update({positionsBySiteId})
    }
    siteDidJoin(siteId) {
        this.joinEvents.push(siteId);
    }
    siteDidLeave(siteId) {
        this.leaveEvents.push(siteId);
    }
    didChangeEditorProxies() { }
}
exports.default = PortalBinding;
//# sourceMappingURL=PortalBinding.js.map